<?php

namespace Drupal\custom_report\Controller;

use Drupal\Core\Controller\ControllerBase;
use Drupal\Core\Datetime\DateFormatterInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\File\FileSystemInterface;
use Drupal\Core\Extension\ModuleHandlerInterface;
use Drupal\Core\Extension\ThemeHandlerInterface;
use Drupal\Core\Messenger\MessengerInterface;

/**
 * Controller for generating custom reports.
 *
 * This controller handles the generation and display of custom reports
 * about core changes, custom modules, and custom themes.
 *
 * @category Controller
 * @package Custom_Report
 * @license GPL-2.0-or-later
 * @link https://www.drupal.org/project/custom_report
 */
class ReportController extends ControllerBase {

  /**
   * The date formatter service.
   *
   * @var \Drupal\Core\Datetime\DateFormatterInterface
   */
  protected $dateFormatter;

  /**
   * The entity type manager.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;

  /**
   * The file system service.
   *
   * @var \Drupal\Core\File\FileSystemInterface
   */
  protected $fileSystem;

  /**
   * The module handler service.
   *
   * @var \Drupal\Core\Extension\ModuleHandlerInterface
   */
  protected $moduleHandler;

  /**
   * The theme handler service.
   *
   * @var \Drupal\Core\Extension\ThemeHandlerInterface
   */
  protected $themeHandler;

  /**
   * The messenger service.
   *
   * @var \Drupal\Core\Messenger\MessengerInterface
   */
  protected $messenger;

  /**
   * The app root.
   *
   * @var string
   */
  protected $appRoot;

  /**
   * Constructs a ReportController object.
   *
   * @param \Drupal\Core\Datetime\DateFormatterInterface $date_formatter
   *   The date formatter service.
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager
   *   The entity type manager.
   * @param \Drupal\Core\File\FileSystemInterface $file_system
   *   The file system service.
   * @param \Drupal\Core\Extension\ModuleHandlerInterface $module_handler
   *   The module handler service.
   * @param \Drupal\Core\Extension\ThemeHandlerInterface $theme_handler
   *   The theme handler service.
   * @param \Drupal\Core\Messenger\MessengerInterface $messenger
   *   The messenger service.
   */
  public function __construct(
  DateFormatterInterface $date_formatter,
  EntityTypeManagerInterface $entity_type_manager,
  FileSystemInterface $file_system,
  ModuleHandlerInterface $module_handler,
  ThemeHandlerInterface $theme_handler,
  MessengerInterface $messenger
  ) {
    $this->dateFormatter     = $date_formatter;
    $this->entityTypeManager = $entity_type_manager;
    $this->fileSystem        = $file_system;
    $this->moduleHandler     = $module_handler;
    $this->themeHandler      = $theme_handler;
    $this->messenger         = $messenger;
    $this->appRoot           = DRUPAL_ROOT;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    return new static(
    $container->get('date.formatter'),
    $container->get('entity_type.manager'),
    $container->get('file_system'),
    $container->get('module_handler'),
    $container->get('theme_handler'),
    $container->get('messenger')
    );
  }

  /**
   * Generates a custom report.
   *
   * @return array
   *   A render array for the report page.
   */
  public function generateReport() {
    // Get core changes.
    $core_changes = $this->getCoreChanges();

    // Get custom modules.
    $custom_modules = $this->getCustomModules();

    // Get custom themes.
    $custom_themes = $this->getCustomThemes();

    // Build the render array.
    $build = [
      '#theme'          => 'custom_report_display',
      '#title'          => $this->t('Site Structure Report'),
      '#core_changes'   => $core_changes,
      '#custom_modules' => $custom_modules,
      '#custom_themes'  => $custom_themes,
      '#attached'       => [
        'library' => [
          'custom_report/report_styles',
        ],
      ],
      '#cache'          => [
        'max-age' => 3600,
      ],
    ];

    return $build;
  }

  /**
   * Get list of modified core files.
   *
   * @return array
   *   Array of modified core files.
   */
  protected function getCoreChanges() {
    $modified_files = [];
    $drupal_root    = $this->appRoot;

    // Path to core directories to check.
    $core_paths = [
      'core/includes',
      'core/lib',
      'core/misc',
      'core/modules',
      'core/profiles',
      'core/themes',
      'core/assets',
      'core/scripts',
      'core/tests',
    ];

    // Get Drupal version from Drupal class.
    $drupal_version    = '';
    $version_file_path = $drupal_root . '/core/lib/Drupal.php';
    if (file_exists($version_file_path)) {
      $version_content = file_get_contents($version_file_path);
      if (preg_match("/const VERSION = '([^']+)'/", $version_content, $matches)) {
        $drupal_version = $matches[1];
      }
    }

    // If we couldn't detect the version, try composer.lock.
    if (empty($drupal_version)) {
      $composer_lock_path = dirname($drupal_root) . '/composer.lock';
      if (file_exists($composer_lock_path)) {
        $composer_lock = json_decode(file_get_contents($composer_lock_path), TRUE);
        if (!empty($composer_lock['packages'])) {
          foreach ($composer_lock['packages'] as $package) {
            if ($package['name'] === 'drupal/core' || $package['name'] === 'drupal/core-recommended') {
              $drupal_version = ltrim($package['version'], 'v');
              break;
            }
          }
        }
      }
    }

    // Create a temporary directory to clone the Drupal Git repository.
    $temp_dir = sys_get_temp_dir() . '/drupal_core_' . md5(microtime());
    $this->fileSystem->prepareDirectory($temp_dir, FileSystemInterface::CREATE_DIRECTORY);

    // Set up the message about which version we're checking against.
    $version_message = '';
    if (!empty($drupal_version)) {
      $version_message = $this->t(
      'Checking against Drupal @version core files',
      ['@version' => $drupal_version]
      );
      $this->messenger->addStatus($version_message);
    }
    else {
      $this->messenger->addWarning(
      $this->t(
       'Could not determine Drupal version. Using latest core files for comparison.'
      )
         );
    }

    // Get Git repository information if available and
    // clone the reference core repo.
    $has_reference_repo = FALSE;
    try {
      // Clone only the core directory, with depth 1 to minimize download size.
      $git_tag       = !empty($drupal_version) ? $drupal_version : 'latest';
      $clone_command = sprintf(
      'git clone --depth 1 --branch %s --single-branch https://git.drupalcode.org/project/drupal.git %s 2>&1',
      escapeshellarg($git_tag),
      escapeshellarg($temp_dir)
      );

      // Execute the clone command.
      exec($clone_command, $clone_output, $clone_return_var);

      if ($clone_return_var === 0) {
        $has_reference_repo = TRUE;
      }
      else {
        /* If the specific version tag failed, try with the major
         *  version tag (e.g., 10.0.x).
         */
        if (!empty($drupal_version)) {
          $major_version_parts = explode('.', $drupal_version);
          if (count($major_version_parts) >= 2) {
            $major_version_branch = $major_version_parts[0] . '.' . $major_version_parts[1] . '.x';

            // Clean up failed clone attempt.
            $this->fileSystem->deleteRecursive($temp_dir);
            $this->fileSystem->prepareDirectory($temp_dir, FileSystemInterface::CREATE_DIRECTORY);

            $clone_command = sprintf(
            'git clone --depth 1 --branch %s --single-branch https://git.drupalcode.org/project/drupal.git %s 2>&1',
            escapeshellarg($major_version_branch),
            escapeshellarg($temp_dir)
            );

            exec($clone_command, $clone_output, $clone_return_var);

            if ($clone_return_var === 0) {
              $has_reference_repo = TRUE;
              $this->messenger->addStatus($this->t('Using Drupal @branch branch for comparison', ['@branch' => $major_version_branch]));
            }
          }
        }

        // If still failed, try with the default branch.
        if (!$has_reference_repo) {
          // Clean up failed clone attempt.
          $this->fileSystem->deleteRecursive($temp_dir);
          $this->fileSystem->prepareDirectory($temp_dir, FileSystemInterface::CREATE_DIRECTORY);

          $clone_command = 'git clone --depth 1 https://git.drupalcode.org/project/drupal.git ' . escapeshellarg($temp_dir) . ' 2>&1';
          exec($clone_command, $clone_output, $clone_return_var);

          if ($clone_return_var === 0) {
            $has_reference_repo = TRUE;
            $this->messenger->addStatus($this->t('Using Drupal default branch for comparison'));
          }
          else {
            $this->messenger->addError(
              $this->t(
               'Failed to clone Drupal Git repository. Error: @error',
               [
                 '@error' => implode("\n", $clone_output),
               ]
              )
                       );
          }
        }
      }

      // If we have the reference repository, compare files.
      if ($has_reference_repo) {
        // Loop through each core directory.
        foreach ($core_paths as $core_path) {
          $this->compareDirectoryWithReference(
          $drupal_root . '/' . $core_path,
          $temp_dir . '/' . $core_path,
          $drupal_root,
          $modified_files
          );
        }

        // Also check the core root directory.
        $this->compareDirectoryWithReference(
        $drupal_root . '/core',
        $temp_dir . '/core',
        $drupal_root,
        $modified_files,
        1
        );
      }

      // Clean up temporary directory.
      $this->fileSystem->deleteRecursive($temp_dir);
    }
    catch (\Exception $e) {
      $this->messenger->addError(
      $this->t(
       'Error comparing core files: @error',
       [
         '@error' => $e->getMessage(),
       ]
      )
         );

      // Try to clean up temp directory in case of failure.
      if (file_exists($temp_dir)) {
        $this->fileSystem->deleteRecursive($temp_dir);
      }
    }

    // Fallback to existing method if Git reference checkout failed.
    if (!$has_reference_repo) {
      $this->messenger->addWarning(
      $this->t(
       'Could not use Git for comparison. Falling back to timestamp-based detection method.'
      )
      );

      // Get baseline timestamp for comparison.
      $baseline_timestamp = 0;

      // Find a reliable baseline timestamp
      // - core.api.php is typically not modified.
      if (file_exists($drupal_root . '/core/core.api.php')) {
        $baseline_timestamp = filemtime($drupal_root . '/core/core.api.php');
      }
      // If we can't get that, try the .htaccess file.
      elseif (file_exists($drupal_root . '/.htaccess')) {
        $baseline_timestamp = filemtime($drupal_root . '/.htaccess');
      }

      // Check local Git repository if available.
      $has_git    = FALSE;
      $git_output = [];
      if (is_dir($drupal_root . '/.git')) {
        $has_git = TRUE;
        // Execute git command to get modified core files.
        $command = 'cd ' . escapeshellarg($drupal_root) . ' && git diff --name-only HEAD';
        exec($command, $git_output, $return_var);

        if ($return_var === 0) {
          foreach ($git_output as $file) {
            // Check if this is a core file.
            if (strpos($file, 'core/') === 0) {
              $modified_files[] = [
                'filename'         => basename($file),
                'filepath'         => $file,
                'modified_date'    => filemtime($drupal_root . '/' . $file),
                'detection_method' => 'git-diff',
              ];
            }
          }
        }

        // Also check for untracked and unstaged changes.
        $command         = 'cd ' . escapeshellarg($drupal_root) . ' && git ls-files --modified --others --exclude-standard';
        $unstaged_output = [];
        exec($command, $unstaged_output, $return_var);

        if ($return_var === 0) {
          foreach ($unstaged_output as $file) {
            // Check if this is a core file.
            if (strpos($file, 'core/') === 0) {
              $full_path = $drupal_root . '/' . $file;
              if (file_exists($full_path)) {
                $modified_files[] = [
                  'filename'         => basename($file),
                  'filepath'         => $file,
                  'modified_date'    => filemtime($full_path),
                  'detection_method' => 'git-unstaged',
                ];
              }
            }
          }
        }
      }

      // Scan all core directories for modified files regardless of Git.
      if ($baseline_timestamp > 0) {
        foreach ($core_paths as $core_path) {
          $this->scanDirectoryForModifiedFiles($drupal_root . '/' . $core_path, $drupal_root, $baseline_timestamp, $modified_files);
        }

        // Also check core files in the root core directory.
        $this->scanDirectoryForModifiedFiles($drupal_root . '/core', $drupal_root, $baseline_timestamp, $modified_files, 1);
      }
      else {
        // No reliable baseline - add a message.
        $this->messenger->addWarning(
          $this->t(
           'Could not determine a reliable baseline for detecting core changes. For accurate detection of core changes, a version control system is recommended.'
          )
               );
      }
    }

    // Special case for batch.inc - if you specifically modified this file.
    $batch_file           = 'core/includes/batch.inc';
    $batch_file_full_path = $drupal_root . '/' . $batch_file;

    if (file_exists($batch_file_full_path)) {
      $is_duplicate = FALSE;

      // Check if batch.inc is already in our list.
      foreach ($modified_files as $existing) {
        if ($existing['filepath'] == $batch_file) {
          $is_duplicate = TRUE;
          break;
        }
      }
    }

    // Sort by modified date, most recent first.
    if (!empty($modified_files)) {
      usort(
      $modified_files,
      function ($a, $b) {
        return $b['modified_date'] <=> $a['modified_date'];
      }
      );

      // Format dates.
      foreach ($modified_files as &$file) {
        $file['modified_date'] = $this->dateFormatter->format($file['modified_date'], 'medium');
      }
    }

    return $modified_files;
  }

  /**
   * Compare files between installed Drupal and reference repository.
   *
   * @param string $site_dir
   *   The directory path in the site to compare.
   * @param string $ref_dir
   *   The directory path in the reference repo to compare against.
   * @param string $drupal_root
   *   The Drupal root path.
   * @param array &$modified_files
   *   Array to store found modified files.
   * @param int $max_depth
   *   Maximum depth to recursively scan (default: 10).
   * @param int $current_depth
   *   Current depth in the recursion (used internally).
   */
  protected function compareDirectoryWithReference($site_dir, $ref_dir, $drupal_root, array &$modified_files, $max_depth = 10, $current_depth = 0) {
    if (!is_dir($site_dir) || !is_dir($ref_dir) || $current_depth > $max_depth) {
      return;
    }

    // Get files in the site directory.
    $site_files = scandir($site_dir);
    foreach ($site_files as $file) {
      if ($file === '.' || $file === '..') {
        continue;
      }

      $site_path     = $site_dir . '/' . $file;
      $ref_path      = $ref_dir . '/' . $file;
      $relative_path = str_replace($drupal_root . '/', '', $site_path);

      if (is_dir($site_path)) {
        // Recursively scan subdirectories.
        if (is_dir($ref_path)) {
          $this->compareDirectoryWithReference($site_path, $ref_path, $drupal_root, $modified_files, $max_depth, $current_depth + 1);
        }
        else {
          // Directory exists in site but not in reference - could be custom.
          $this->scanDirectoryForFiles($site_path, $drupal_root, $modified_files, 'new-directory', $max_depth, $current_depth + 1);
        }
      }
      elseif (is_file($site_path)) {
        // Check if this is a PHP, JS, CSS, or other important file type.
        if (preg_match('/\.(php|inc|module|install|yml|js|css|twig|html|json)$/i', $file)) {
          // If the file doesn't exist in reference or content differs.
          if (!file_exists($ref_path) || !$this->areFilesIdentical($site_path, $ref_path)) {
            $is_duplicate = FALSE;

            // Check if this file is already in the list.
            foreach ($modified_files as $existing) {
              if ($existing['filepath'] == $relative_path) {
                $is_duplicate = TRUE;
                break;
              }
            }

            if (!$is_duplicate) {
              $modified_files[] = [
                'filename'         => basename($site_path),
                'filepath'         => $relative_path,
                'modified_date'    => filemtime($site_path),
                'detection_method' => file_exists($ref_path) ? 'content-diff' : 'new-file',
              ];
            }
          }
        }
      }
    }
  }

  /**
   * Check if two files have identical content.
   *
   * @param string $file1
   *   Path to the first file.
   * @param string $file2
   *   Path to the second file.
   *
   * @return bool
   *   TRUE if files are identical, FALSE otherwise.
   */
  protected function areFilesIdentical($file1, $file2) {
    // Quick size check first.
    if (filesize($file1) !== filesize($file2)) {
      return FALSE;
    }

    // For small files, compare content directly.
    // 1MB limit for direct comparison.
    if (filesize($file1) < 1024 * 1024) {
      return file_get_contents($file1) === file_get_contents($file2);
    }

    // For larger files, compare hashes.
    return md5_file($file1) === md5_file($file2);
  }

  /**
   * Scan a directory for all files recursively.
   *
   * @param string $dir
   *   The directory to scan.
   * @param string $drupal_root
   *   The Drupal root path.
   * @param array &$modified_files
   *   Array to store found files.
   * @param string $detection_method
   *   Method used to detect these files.
   * @param int $max_depth
   *   Maximum depth to recursively scan (default: 10).
   * @param int $current_depth
   *   Current depth in the recursion (used internally).
   */
  protected function scanDirectoryForFiles($dir, $drupal_root, array &$modified_files, $detection_method = 'scan', $max_depth = 10, $current_depth = 0) {
    if (!is_dir($dir) || $current_depth > $max_depth) {
      return;
    }

    $files = scandir($dir);

    foreach ($files as $file) {
      if ($file === '.' || $file === '..') {
        continue;
      }

      $path          = $dir . '/' . $file;
      $relative_path = str_replace($drupal_root . '/', '', $path);

      if (is_dir($path)) {
        // Recursively scan subdirectories.
        $this->scanDirectoryForFiles($path, $drupal_root, $modified_files, $detection_method, $max_depth, $current_depth + 1);
      }
      elseif (is_file($path)) {
        // Check if this is a PHP, JS, CSS, or other important file type.
        if (preg_match('/\.(php|inc|module|install|yml|js|css|twig|html|json)$/i', $file)) {
          // Add to the files list unless it's already there.
          $is_duplicate = FALSE;

          // Check if this file is already in the list.
          foreach ($modified_files as $existing) {
            if ($existing['filepath'] == $relative_path) {
              $is_duplicate = TRUE;
              break;
            }
          }

          if (!$is_duplicate) {
            $modified_files[] = [
              'filename'         => basename($path),
              'filepath'         => $relative_path,
              'modified_date'    => filemtime($path),
              'detection_method' => $detection_method,
            ];
          }
        }
      }
    }
  }

  /**
   * Get list of custom modules.
   *
   * @return array
   *   Array of custom modules.
   */
  protected function getCustomModules() {
    $custom_modules = [];
    $modules        = $this->moduleHandler->getModuleList();

    foreach ($modules as $name => $module) {
      $path = $module->getPath();

      // Check if this is a custom module (in web/modules/custom).
      if (strpos($path, 'modules/custom') !== FALSE) {
        // Safely get module info.
        try {
          $module_info = $this->moduleHandler->getModule($name);

          // Safer check for module info -
          // first check if the property exists
          // and is accessible.
          $info = [];

          // Get the module info from the extension
          // info file directly if needed.
          if (method_exists($module_info, 'getInfo')) {
            $info = $module_info->getInfo();
          }
          elseif (isset($module_info->info)) {
            $info = $module_info->info;
          }

          // Always provide fallback values using null coalescing operator.
          $custom_modules[] = [
            'name'        => $name,
            'path'        => $path,
            'title'       => $info['name'] ?? $name,
            'description' => $info['description'] ?? $this->t('No description available'),
            'version'     => $info['version'] ?? '',
            'status'      => $this->moduleHandler->moduleExists($name) ? $this->t('Enabled') : $this->t('Disabled'),
          ];
        }
        catch (\Exception $e) {
          // Add basic info if there was an exception.
          $custom_modules[] = [
            'name'        => $name,
            'path'        => $path,
            'title'       => $name,
            'description' => $this->t(
           'Error getting module information: @error',
           ['@error' => $e->getMessage()]
            ),
            'version'     => '',
            'status'      => $this->moduleHandler->moduleExists($name)
              ? $this->t('Enabled')
              : $this->t('Disabled'),
          ];
        }
      }
    }

    // Sort alphabetically by title.
    usort(
    $custom_modules,
    function ($a, $b) {
      return strcasecmp($a['title'], $b['title']);
    }
    );

    return $custom_modules;
  }

  /**
   * Get list of custom themes.
   *
   * @return array
   *   Array of custom themes.
   */
  protected function getCustomThemes() {
    $custom_themes = [];
    $themes        = $this->themeHandler->listInfo();

    foreach ($themes as $name => $theme) {
      $path = $theme->getPath();

      // Check if this is a custom theme (in web/themes/custom).
      if (strpos($path, 'themes/custom') !== FALSE) {
        // Safely get theme info.
        try {
          $theme_info = $this->themeHandler->getTheme($name);

          // Safer check for theme info.
          $info = [];

          // Get the theme info from the extension info file directly if needed.
          if (method_exists($theme_info, 'getInfo')) {
            $info = $theme_info->getInfo();
          }
          elseif (isset($theme_info->info)) {
            $info = $theme_info->info;
          }

          $status = $this->themeHandler->themeExists($name) ?
          ($theme->status ? $this->t('Enabled') : $this->t('Disabled')) :
          $this->t('Not installed');

          // Always provide fallback values.
          $custom_themes[] = [
            'name'        => $name,
            'path'        => $path,
            'title'       => $info['name'] ?? $name,
            'description' => $info['description'] ?? $this->t('No description available'),
            'version'     => $info['version'] ?? '',
            'status'      => $status,
          ];
        }
        catch (\Exception $e) {
          // Add basic info if there was an exception.
          $custom_themes[] = [
            'name'        => $name,
            'path'        => $path,
            'title'       => $name,
            'description' => $this->t(
           'Error getting theme information: @error',
           [
             '@error' => $e->getMessage(),
           ]
            ),
            'version'     => '',
            'status'      => $this->themeHandler->themeExists($name) ?
            ($theme->status ? $this->t('Enabled') : $this->t('Disabled')) :
            $this->t('Not installed'),
          ];
        }
      }
    }

    // Sort alphabetically by title.
    usort(
    $custom_themes,
    function ($a, $b) {
      return strcasecmp($a['title'], $b['title']);
    }
    );

    return $custom_themes;
  }

}
